import re
import time
import random
import sys
import socket
import traceback as tback
import threading

from . import specs, parsing


class FreRuntimeError(BaseException):
    pass

class FreMissingError(FreRuntimeError):
    pass

class FreSocketError(FreRuntimeError):
    pass

class FreNameError(FreRuntimeError):
    pass

class FreTypeError(FreRuntimeError):
    pass

class FreValueError(FreRuntimeError):
    pass


class FreVariable(object):
    def __init__(self, env, kind, name, value = None):
        self.env = env # can be either a FreVM or FreEnvironment object

        self.kind = kind
        self.name = name

        if not isinstance(kind, specs.LanguageType):
            raise TypeError("{} is not a fre.specs.LanguageType subclass!".format(type(kind)))

        self.set(value)

        env.variables[name] = self

    def set(self, value):
        if value is None:
            value = self.kind.default_value

        elif not self.kind.check(value):
            raise FreTypeError("{} is not a {}-like value!".format(repr(value), str(self.kind)))

        self.value = self.kind.autocast(value)

    def pointer_level(self):
        return 0

    def get(self):
        return self.value

    def _fre_deref(self):
        if not isinstance(self.kind, specs.PointerType):
            raise FreTypeError("{} can't be dereferenced!".format(self.kind))

        return self.value._fre_deref()

    def pointer(self):
        return FrePointer(self.env, self.kind, self.name, self)

class FreConstant(FreVariable):
    def __init__(self, env, kind, name, value = None):
        self.env = env # can be either a FreVM or FreEnvironment object

        self.kind = kind
        self.name = name

        if not isinstance(kind, specs.LanguageType):
            raise TypeError("{} is not a fre.specs.LanguageType subclass!".format(type(kind)))

        assert self.kind.check(value)
        self.value = value

        env.variables[name] = self

    def set(self, value):
        raise FreRuntimeError("Cannot set a constant variable ({})!".format(self.name))

class FrePointer(object):
    def __init__(self, env, kind, name, var = None):
        self.env = env # can be either a FreVM or FreEnvironment object

        self.kind = kind
        self.name = name

        if not isinstance(kind, specs.LanguageType):
            raise FreTypeError("{} is not a fre.specs.LanguageType subclass".format(type(kind)))

        self.set(var)

        env.variables[name] = self

    def set(self, var):
        if var is None:
            self.value = None

        elif hash(var.kind) != hash(self.kind):
            raise FreTypeError("Pointer assignment mismatch: expected {}*, got {}*".format(str(self.kind), str(var.kind)))

        self.value = var

    def pointer_level(self):
        return self.value.pointer_level() + 1

    def get(self):
        return self

    def _fre_deref(self):
        if self.value is None:
            raise FreValueError("{} is a null pointer".format(self.name))

        else:
            return self.value

class FreFunction(object):
    def __init__(self, env, name, return_kind, arg_typenames, body):
        self.env = env
        self._parser = None

        self.name = name
        self.return_kind = return_kind
        self.arg_typenames = list(arg_typenames)
        self.body = body

        if not isinstance(return_kind, specs.LanguageType):
            raise TypeError("{} is not a fre.specs.LanguageType subclass".format(type(return_kind)))

    def get(self):
        return self

    def call(self, caller, call_env, traceback, args, parser = None):
        call_env.bind(self.env)
        return call_env.execute([self.body], traceback, parser)

    def __mul__(self, num):
        return FreRepeatFunction(self, num)

class FreNativeFunction(FreFunction):
    def __init__(self, env, name, return_kind, arg_typenames):
        self.env = env
        self._parser = None

        self.name = name
        self.return_kind = return_kind
        self.arg_typenames = list(arg_typenames)

        if not isinstance(return_kind, specs.LanguageType):
            raise TypeError("{} is not a fre.specs.LanguageType subclass!".format(type(return_kind)))

    def call(self, caller, call, traceback, args):
        pass

    def __mul__(self, n):
        return FreRepeatFunction(self, n)

class FreRepeatFunction(FreNativeFunction):
    def __init__(self, func, times):
        self.env = func.env

        self.name = func.name + '(*{})'.format(times)
        self.return_kind = func.return_kind
        self.arg_typenames = list(func.arg_typenames)

        self.func = func
        self.times = times

        if not isinstance(func.return_kind, specs.LanguageType):
            raise TypeError("{} is not a fre.specs.LanguageType subclass!".format(type(return_kind)))

    def call(self, caller, call, traceback, args):
        res = None

        for _ in range(self.times):
            res = self.func.call(caller, call, traceback, args)

        return res

    def __mul__(self, n):
        return FreRepeatFunction(self.func, times * n)

#====================
#  NATIVE FUNCTIONS 
#====================

class FreNativeSuite(object):
    operations = []
    constants = {}

    @classmethod
    def register(cls, env):
        for op in cls.operations:
            env.variables[op.fname] = op(env, op.fname, op.rtype, op.atype)

        for cname, (ckind, cval) in cls.constants:
            FreConstant(env, ckind, cname, cval)

# General
class FrePrintFunction(FreNativeFunction):
    def call(self, caller, call, traceback, args):
        sys.stdout.write(str(args[0]))

class FreExitFunction(FreNativeFunction):
    def exit(self, caller, call, traceback, args):
        exit(0)

# Arrays
class FreInsertFunction(FreNativeFunction):
    def call(self, caller, call, traceback, args):
        (array, index, item) = args

        array.insert(index, item)
        return item

class FreAppendFunction(FreNativeFunction):
    def call(self, caller, call, traceback, args):
        array = args[0]
        i = len(array)

        array.append(args[1])

        return i

class FreConcatFunction(FreNativeFunction):
    def call(self, caller, call, traceback, args):
        return list(args[0]) + list(args[1])

class FrePopFunction(FreNativeFunction):
    def call(self, caller, call, traceback, args):
        array = args[0]
        return array.pop()

class FreShiftFunction(FreNativeFunction):
    def call(self, caller, call, traceback, args):
        array = args[0]
        return array.pop(0)

class FreUnshiftFunction(FreNativeFunction):
    def call(self, caller, call, traceback, args):
        array = args[0]
        return array.insert(0, args[1])

class FreArrayGetFunction(FreNativeFunction):
    def call(self, caller, call, traceback, args):
        array = args[0]
        return array[args[1]]

class FreArraySetFunction(FreNativeFunction):
    def call(self, caller, call, traceback, args):
        array = args[0]
        array[args[1]] = args[2]
        return args[2]

# Time

class FreTime(FreNativeSuite):
    class Timestamp(FreNativeFunction):
        fname = 'time:stamp'

        atype = []
        rtype = specs.FloatType()

        def call(self, *args):
            return time.time()

    class Sleep(FreNativeFunction):
        fname = 'time:sleep'

        atype = [(specs.FloatType(), 'seconds')]
        rtype = specs.VoidType()

        def call(self, _, _1, _2, args):
            time.sleep(args[0])

    class Format(FreNativeFunction):
        fname = 'time:format'

        atype = [
            (specs.StringType(), 'format') # format
        ]
        rtype = specs.StringType()

        def call(self, _, _1, _2, args):
            return time.strftime(args[0])

    operations = [
        Timestamp, Format, Sleep
    ]

# Random

class FreRandom(FreNativeSuite):
    rng = random.Random()
    rng.seed(time.time())

    class Integer(FreNativeFunction):
        fname = 'random:int'

        atype = [
            (specs.IntType(), 'min'), # smallest value (inclusive)
            (specs.IntType(), 'max'), # largest value (inclusive)
        ]
        rtype = specs.IntType()

        def call(self, caller, call, traceback, args):
            return FreRandom.rng.randint(args[0], args[1])

    class Uniform(FreNativeFunction):
        fname = 'random:uniform'

        atype = [
            (specs.FloatType(), 'min'), # smallest value
            (specs.FloatType(), 'max'), # largest value
        ]
        rtype = specs.FloatType()

        def call(self, caller, call, traceback, args):
            return FreRandom.rng.uniform(args[0], args[1])

    class Gaussian(FreNativeFunction):
        fname = 'random:gaussian'

        atype = [
            (specs.FloatType(), 'min'), # smallest value
            (specs.FloatType(), 'max'), # largest value
        ]
        rtype = specs.FloatType()

        def call(self, caller, call, traceback, args):
            return FreRandom.rng.gauss(args[0], args[1])

    class Seed(FreNativeFunction):
        fname = 'random:seed'

        atype = [
            (specs.IntType(), 'seed') # seed
        ]
        rtype = specs.VoidType()

        def call(self, caller, call, traceback, args):
            FreRandom.rng.seed(args[0])

    operations = [
        Integer, Uniform, Gaussian,
        Seed
    ]

# Strings

class FreStringUtils(FreNativeSuite):
    class Format(FreNativeFunction):
        fname = 'str:format'
        
        atype = [
            (specs.StringType(), 'format'), # to format
            (specs.ArrayType(specs.AnyType()), 'params') # format parameters
        ]

        rtype = specs.StringType() # formatted

        def call(self, caller, call, traceback, args):
            return args[0].format(*args[1])

    class Slice(FreNativeFunction):
        fname = 'str:slice'
        
        atype = [
            (specs.StringType(), 'source'), # to slice
            (specs.IntType(), 'from'), # index 1
            (specs.IntType(), 'to'), # index 2
        ]

        rtype = specs.StringType() # sliced

        def call(self, caller, call, traceback, args):
            return args[0][args[1]:args[2]]

    class SliceFrom(FreNativeFunction):
        fname = 'str:sliceFrom'
        
        atype = [
            (specs.StringType(), 'source'), # to slice
            (specs.IntType(), 'from'), # index 1
        ]

        rtype = specs.StringType() # sliced

        def call(self, caller, call, traceback, args):
            return args[0][args[1]:]

    class SliceTo(FreNativeFunction):
        fname = 'str:sliceTo'
        
        atype = [
            (specs.StringType(), 'source'), # to slice
            (specs.IntType(), 'to'), # index 2
        ]

        rtype = specs.StringType() # sliced

        def call(self, caller, call, traceback, args):
            return args[0][:args[1]]

    class CharAt(FreNativeFunction):
        fname = 'str:charAt'

        atype = [
            (specs.StringType(), 'source'),
            (specs.IntType(), 'index'),
        ]

        rtype = specs.StringType() # char

        def call(self, caller, call, traceback, args):
            return args[0][args[1]]

    class SetChar(FreNativeFunction):
        fname = 'str:setChar'

        atype = [
            (specs.StringType(), 'target'),
            (specs.IntType(), 'index'), # index
            (specs.StringType(), 'char') # new char
        ]

        rtype = specs.StringType() # with set char

        def call(self, caller, call, traceback, args):
            a = args[0]
            a[args[1]] = args[2][0]
            return a

    class Replace(FreNativeFunction):
        fname = 'str:replace'

        atype = [
            (specs.StringType(), 'source'), # source string
            (specs.StringType(), 'from'), # replace this
            (specs.StringType(), 'to') # by this
        ]

        rtype = specs.StringType() # after replacing

        def call(self, caller, call, traceback, args):
            return args[0].replace(args[1], args[2])

    operations = [
        Format,
        SliceFrom, SliceTo, Slice,
        CharAt, SetChar, Replace
    ]

# Threading

class FreThreads(FreNativeSuite):
    threads = {}
    handle_count = 0

    def regthread(th):
        ind = FreThreads.handle_count
        FreThreads.handle_count += 1

        FreThreads.threads[ind] = th

        return ind

    def _target(source, fre_func, call_env, arg):
        fre_func.call(None, call_env, [('thread start', None, None)], [arg], source._parser)

    class Run(FreNativeFunction):
        fname = 'thread:run'
        
        atype = [
            (specs.FunctionType([], specs.VoidType()), 'target'), # function to be threaded
        ]

        rtype = specs.IntType() # thread handle

        def call(self, caller, call, traceback, args):
            th = threading.Thread(target=FreThreads._target, args=(self, args[0], call, None))
            th.start()

            return FreThreads.regthread(th)

    class RunWith(FreNativeFunction):
        fname = 'thread:runWith'
        
        atype = [
            (specs.FunctionType([specs.AnyType()], specs.VoidType()), 'target'), # function to be threaded
            (specs.AnyType(), 'arg'), # sole argument (may be null)
        ]

        rtype = specs.IntType() # thread handle

        def call(self, caller, call, traceback, args):
            th = threading.Thread(target=FreThreads._target, args=(self, args[0], call, args[1]))
            th.start()

            return FreThreads.regthread(th)

    class Join(FreNativeFunction):
        fname = 'thread:join'

        atype = [
            (specs.IntType(), 'handle') # thread handle
        ]

        rtype = specs.VoidType()

        def call(self, caller, call, traceback, args):
            ind = args[0]

            if ind not in FreThreads.threads:
                raise FreMissingError("No such thread with handle {}!".format(ind))

            th = FreThreads.threads[ind]
            th.join()

    operations = [
        Run, RunWith,
        Join,
    ]

# Networking

class FreSocket(FreNativeSuite):
    sockets = {}
    addresses = {}
    handle_count = 0
    modes = [socket.SOCK_STREAM, socket.SOCK_DGRAM]

    class Open(FreNativeFunction):
        fname = 'socket:open'

        atype = [
            (specs.IntType(), "mode"), # SOCK_TCP or SOCK_UDP
        ]

        rtype = specs.IntType() # socket handle

        def call(self, caller, call, traceback, args):
            sock = socket.socket(socket.AF_INET, FreSocket.modes[args[0]])
            return FreSocket.regsock(sock, None)

    class Close(FreNativeFunction):
        fname = 'socket:close'
        
        atype = [
            (specs.IntType(), "socket"), # socket handle
        ]

        rtype = specs.VoidType()

        def call(self, caller, call, traceback, args):
            ind = args[0]

            if ind not in FreSocket.sockets:
                raise FreMissingError("No socket under handle {}!".format(ind))

            sock = FreSocket.sockets[ind]
            sock.close()

            del FreSocket.sockets[ind]

            if ind in FreSocket.addresses:
                del FreSocket.addresses[ind]

    class Connect(FreNativeFunction):
        fname = 'socket:connect'

        atype = [
            (specs.IntType(), "socket") , # socket handle
            (specs.StringType(), "host"), # 1st part of AF_INET address
            (specs.IntType(), "port"),    # 2nd part of AF_INET address
        ]

        rtype = specs.VoidType()

        def call(self, caller, call, traceback, args):
            ind = args[0]

            if ind not in FreSocket.sockets:
                raise FreMissingError("No socket under handle {}!".format(ind))

            sock = FreSocket.sockets[ind]

            try:
                sock.connect((args[1], args[2]))

            except socket.error as err:
                raise FreSocketError(str(err))

    class Send(FreNativeFunction):
        fname = 'socket:send'

        atype = [
            (specs.IntType(), "socket") , # socket handle
            (specs.StringType(), "data"),
        ]

        rtype = specs.VoidType()

        def call(self, caller, call, traceback, args):
            ind = args[0]

            if ind not in FreSocket.sockets:
                raise FreMissingError("No socket under handle {}!".format(ind))

            sock = FreSocket.sockets[ind]

            msg = args[1]

            if isinstance(msg, str):
                msg = msg.encode('utf-8')

            try:
                sock.sendall(msg)

            except socket.error as err:
                raise FreSocketError(str(err))

    class ReceiveTCP(FreNativeFunction):
        fname = 'socket:receive'

        atype = [
            (specs.IntType(), "socket"),  # socket handle
            (specs.IntType(), "max_size"),
        ]

        rtype = specs.StringType()

        def call(self, caller, call, traceback, args):
            ind = args[0]

            if ind not in FreSocket.sockets:
                raise FreMissingError("No socket under handle {}!".format(ind))

            sock = FreSocket.sockets[ind]

            try:
                return sock.recv(args[1]).decode('utf-8')

            except socket.error as err:
                raise FreSocketError(str(err))

    class ReceiveUDP(FreNativeFunction):
        fname = 'socket:receiveFrom'

        atype = [
            (specs.IntType(), "socket"),  # socket handle
            (specs.IntType(), "max_size"),
        ]

        rtype = specs.ArrayType(specs.StringType()) # [address, data]

        def call(self, caller, call, traceback, args):
            ind = args[0]

            if ind not in FreSocket.sockets:
                raise FreMissingError("No socket under handle {}!".format(ind))

            sock = FreSocket.sockets[ind]

            try:
                return sock.recvfrom(args[1]).decode('utf-8')

            except socket.error as err:
                raise FreSocketError(str(err))

    class Bind(FreNativeFunction):
        fname = 'socket:bind'

        atype = [
            (specs.IntType(), "socket"),  # socket handle
            (specs.StringType(), "host"), # 1st part of AF_INET address
            (specs.IntType(), "port"),    # 2nd part of AF_INET address
        ]

        rtype = specs.VoidType()

        def call(self, caller, call, traceback, args):
            ind = args[0]

            if ind not in FreSocket.sockets:
                raise FreMissingError("No socket under handle {}!".format(ind))

            sock = FreSocket.sockets[ind]

            try:
                sock.bind((args[1], args[2]))

            except socket.error as err:
                raise FreSocketError(str(err))

    class Listen(FreNativeFunction):
        fname = 'socket:listen'

        atype = [
            (specs.IntType(), "socket"),  # socket handle
            (specs.IntType(), "listenLimit"),
        ]

        rtype = specs.VoidType()

        def call(self, caller, call, traceback, args):
            ind = args[0]

            if ind not in FreSocket.sockets:
                raise FreMissingError("No socket under handle {}!".format(ind))

            sock = FreSocket.sockets[ind]

            try:
                sock.listen(args[1])

            except socket.error as err:
                raise FreSocketError(str(err))

    class Accept(FreNativeFunction):
        fname = 'socket:accept'

        atype = [
            (specs.IntType(), "socket")  # socket handle
        ]

        rtype = specs.IntType() # accepted handle

        def call(self, caller, call, traceback, args):
            ind = args[0]

            if ind not in FreSocket.sockets:
                raise FreMissingError("No socket under handle {}!".format(ind))

            sock = FreSocket.sockets[ind]

            try:
                (conn, addr) = sock.accept()
                return FreSocket.regsock(conn, addr)

            except socket.error as err:
                raise FreSocketError(str(err))

    class GetHost(FreNativeFunction):
        fname = 'socket:getHost'

        atype = [
            (specs.IntType(), "socket")  # socket handle
        ]

        rtype = specs.StringType() # address

        def call(self, caller, call, traceback, args):
            ind = args[0]

            if ind not in FreSocket.sockets:
                raise FreMissingError("No socket under handle {}".format(ind))

            if ind not in FreSocket.addresses:
                raise FreMissingError("Can't retrieve the host of this socket")

            return FreSocket.addresses[ind][0]

    class GetPort(FreNativeFunction):
        fname = 'socket:getPort'

        atype = [
            (specs.IntType(), "socket")  # socket handle
        ]

        rtype = specs.IntType() # port

        def call(self, caller, call, traceback, args):
            ind = args[0]

            if ind not in FreSocket.sockets:
                raise FreMissingError("No socket under handle {}!".format(ind))

            if ind not in FreSocket.addresses:
                raise FreMissingError("Can't retrieve the port of this socket")

            return FreSocket.addresses[ind][1]

    @staticmethod
    def regsock(conn, addr):
        ind = FreSocket.handle_count
        FreSocket.handle_count += 1

        FreSocket.sockets[ind] = conn

        if addr:
            FreSocket.addresses[ind] = addr

        return ind

    operations = [
        Open, Close,
        Connect,
        Send, ReceiveTCP, ReceiveUDP,
        Bind, Listen, Accept,
        GetHost, GetPort
    ]

#===================


class FreEnvironment(object):
    def __init__(self, name, spec = None, parent = None):
        # (cetera)
        self.name = name
        self.parent = parent
        self.specs = spec or specs.FreSpecs()

        # Bindings
        self.variables = {}
        self.super_scopes = []

        if parent is None:
            self.init_native_functions()

        # Execution
        self.result = None

        self.return_stop = False
        self.continue_stop = False
        self.break_stop = False

        self._exec_level = 0
        self._parser = None

    def bind(self, super_scope):
        self.super_scopes.append(super_scope)

    def init_native_functions(self):
        self.variables['print'] = FrePrintFunction(self, 'print', specs.VoidType(), [(specs.AnyType(), 'x')])
        self.variables['exit'] = FreExitFunction(self, 'exit', specs.VoidType(), [])

        # Array Functions
        self.variables['array:concat'] = FreConcatFunction(self, 'array:concat', specs.ArrayType(specs.AnyType()), [(specs.ArrayType(specs.AnyType()), 'arr1'), (specs.ArrayType(specs.AnyType()), 'arr2')])

        self.variables['array:insert'] = FreInsertFunction(self, 'array:insert', specs.IntType(), [(specs.ArrayType(specs.AnyType()), 'arr'), (specs.IntType(), 'index'), (specs.AnyType(), 'item')])
        self.variables['array:append'] = FreAppendFunction(self, 'array:append', specs.IntType(), [(specs.ArrayType(specs.AnyType()), 'arr'), (specs.AnyType(), 'item')])
        self.variables['array:unshift'] = FreUnshiftFunction(self, 'array:unshift', specs.AnyType(), [(specs.ArrayType(specs.AnyType()), 'arr'), (specs.AnyType(), 'item')])
        
        self.variables['array:pop'] = FrePopFunction(self, 'array:pop', specs.AnyType(), [(specs.ArrayType(specs.AnyType()), 'arr')])
        self.variables['array:shift'] = FreShiftFunction(self, 'array:shift', specs.AnyType(), [(specs.ArrayType(specs.AnyType()), 'arr')])

        self.variables['array:get'] = FreArrayGetFunction(self, 'array:get', specs.AnyType(), [(specs.ArrayType(specs.AnyType()), 'arr'), (specs.IntType(), 'index')])
        self.variables['array:set'] = FreArraySetFunction(self, 'array:set', specs.AnyType(), [(specs.ArrayType(specs.AnyType()), 'arr'), (specs.IntType(), 'index'), (specs.AnyType(), 'value')])

        # Socket Constants
        self.add_constant('SOCK_TCP', specs.IntType(), 0)
        self.add_constant('SOCK_UDP', specs.IntType(), 1)

        FreSocket.register(self)
        FreThreads.register(self)
        FreStringUtils.register(self)
        FreTime.register(self)
        FreRandom.register(self)

    def add_constant(self, name, kind, value):
        FreConstant(self, kind, name, value)

    def parse_type(self, name):
        if name == '':
            return specs.VoidType()

        elif name[-1] == '*':
            return specs.PointerType(self.parse_type(name[:-1]))

        elif name[-1] == '%':
            return specs.ArrayType(self.parse_type(name[:-1]))

        elif re.match(r'.+?\(.*?\)', name):
            m = re.match(r'(.+?)\((.*?)\)', name)

            return_type = m.group(1)
            arg_types = [x for x in m.group(2).split(', ') if x]
            
            return specs.FunctionType([self.parse_type(a) for a in arg_types], self.parse_type(return_type))

        else:
            return self.specs.find_type(name)()

    def find_variable(self, name):
        if name in self.variables:
            return self.variables[name]

        else:
            for s in [self.parent, *self.super_scopes]:
                if not s:
                    continue

                try:
                    return s.find_variable(name)

                except FreNameError:
                    continue

            raise FreNameError("Variable {} not found in the Fre environment!".format(name))

    def evaluate(self, node, traceback = None):
        if not traceback:
            traceback = []

        if not node:
            return None

        if node.node_type == 'expr.funcall':
            func = self.find_variable(node.value)

            if func is None:
                return None

            if isinstance(func, FreVariable):
                if isinstance(func.kind, specs.FunctionType):
                    n = func.name
                    func = func.value

                    if func is None:
                        raise FreTypeError("{} is null".format(n))

                else:
                    raise FreTypeError("{} is not a function variable".format(func.name))

            call = FreEnvironment('call to {}'.format(func.name), self.specs, self)
            args = []

            for c in node.children:
                args.append(self.evaluate(c, traceback))

            for i, ((argt, argn), argv) in enumerate(zip(func.arg_typenames, args)):
                if not argt.check(argv):
                    raise FreTypeError("Function {} argument #{} type mismatch: expected {}-like value, got {}".format(func.name, i, argt, type(argv).__name__))
                FreVariable(call, argt, argn, argv)

            func._parser = self._parser

            res = func.call(self, call, traceback, args)
            return res

        elif node.node_type.startswith('expr.literal.'):
            return node.value

        elif node.node_type == 'expr.identifier':
            vname, vmods = node.value
            var = self.find_variable(vname)

            for m in vmods:
                if m == '*':
                    var = var._fre_deref()

                elif m == '@':
                    var = FrePointer(self, var.kind, var.name + '*', var)

            return var.get()

        elif node.node_type == 'expr.oper':
            operator = node.value
            operands = [self.evaluate(n, traceback) for n in node.children]

            if operator == '+':
                res = operands[0]

                for o in operands[1:]:
                    if isinstance(o, str) or isinstance(res, str):
                        res = str(res) + str(o)

                    else:
                        res += o

                return res

            elif operator == '-':
                res = operands[0]

                for o in operands[1:]:
                    res -= o

                return res

            elif operator == '*':
                res = operands[0]

                for o in operands[1:]:
                    res *= o

                return res

            elif operator == '/':
                res = operands[0]

                for o in operands[1:]:
                    res /= o

                return res

            elif operator == '==':
                res = operands[0]

                for o in operands[1:]:
                    if o != res:
                        return False

                return True

            elif operator == '!=':
                res = operands[0]

                for o in operands[1:]:
                    if o != res:
                        return True

                return False

            elif operator == '&':
                res = operands[0]

                for o in operands[1:]:
                    res &= o

                return res

            elif operator == '|':
                res = operands[0]

                for o in operands[1:]:
                    res |= o

                return res

            elif operator == '^':
                res = operands[0]

                for o in operands[1:]:
                    res ^= o

                return res

            elif operator == '<<':
                res = operands[0]

                for o in operands[1:]:
                    res <<= o

                return res

            elif operator == '>>':
                res = operands[0]

                for o in operands[1:]:
                    res >>= o

                return res

            elif operator == '~':
                return ~operands[0]

            elif operator == '!':
                return not operands[0]

            elif operator == '&&':
                res = True
                
                for o in operands:
                    res = res and o

                return res

            elif operator == '||':
                res = False
                
                for o in operands:
                    res = res or o

                return res

            elif operator == '^^':
                res = False
                
                for o in operands:
                    res = bool(res) != bool(o)

                return res

            elif operator == '<':
                last = operands[0]

                for o in operands[1:]:
                    if o <= last:
                        return False

                    last = o

                return True

            elif operator == '>':
                last = operands[0]

                for o in operands[1:]:
                    if o >= last:
                        return False

                    last = o

                return True

            elif operator == '<=':
                last = operands[0]

                for o in operands[1:]:
                    if o < last:
                        return False

                    last = o

                return True

            elif operator == '>=':
                last = operands[0]

                for o in operands[1:]:
                    if o > last:
                        return False

                    last = o

                return True

            else:
                raise FreRuntimeError("Unknown operator: {}".format(repr(operator)))

        elif node.node_type == 'expr.set':
            name, mods = node.value
            var = self.find_variable(name)

            for m in mods:
                if m == '*':
                    var = var._fre_deref()

            value = self.evaluate(node.body, traceback)
            var.set(value)
            
            return value

        elif node.node_type == 'expr.func':
            func = FreFunction(self, node.value, self.parse_type(node.body.value), [(self.parse_type(adef.body.value), adef.value) for adef in node.children[1:]], node.children[0])
            self.variables[node.value] = func
            return func

    def execute(self, nodes, traceback, parser = None):
        try:
            self._exec_level += 1

            for node in nodes:
                if not node:
                    continue

                traceback.append((self.name, node.line, node.column))

                try:
                    # Stoppers
                    if self.return_stop:
                        if self._exec_level == 1:
                            self.return_stop = False

                            r = self.result
                    
                            self._exec_level = 0
                            self.result = None

                            return self.result

                        else:
                            self._exec_level -= 1
                            return

                    elif self.continue_stop:
                        self.continue_stop = False
                        self._exec_level -= 1
                        return

                    elif self.break_stop:
                        self._exec_level -= 1
                        return

                    if not node:
                        continue

                    # Run the AST using a plain old elif chain
                    nt = node.node_type

                    if nt == 'vardef':
                        name, mods, value = node.value
                        #node.print_tree()
                        vtype = self.parse_type(node.body.value)
                        level = 0

                        for m in mods:
                            if m == '@':
                                level += 1

                        if level > 0:
                            target = self.find_variable(value)

                            if target.pointer_level() != level:
                                raise FreTypeError("Pointer level mismatch: Expected a {} type, got {}".format(vtype, target.kind + '*'))

                        else:
                            value = self.evaluate(value, traceback)
                            FreVariable(self, vtype, name, value)

                    elif nt == 'block':
                        traceback.pop()
                        self.execute(node.children, traceback)
                        traceback.append((self.name, node.line, node.column))

                    elif nt == 'logic.if':
                        (condition, otherwise) = node.children

                        if self.evaluate(condition, traceback):
                            self.execute([node.body], traceback)

                        elif otherwise:
                            self.execute([otherwise.body], traceback)

                    elif nt == 'logic.while':
                        (condition, otherwise) = node.children

                        if self.evaluate(condition, traceback):
                            while True:
                                self.execute([node.body], traceback)

                                if self.break_stop:
                                    self.break_stop = False
                                    break

                                elif not self.evaluate(condition, traceback):
                                    break

                        elif otherwise:
                            self.execute([otherwise.body], traceback)

                    elif nt == 'continue':
                        self.continue_stop = True

                    elif nt == 'break':
                        self.break_stop = True

                    elif nt == 'return':
                        self.return_stop = True

                        if node.body:
                            self.result = self.evaluate(node.body, traceback)

                    elif nt == 'pass':
                        continue

                    else:
                        self.evaluate(node, traceback)

                except BaseException as err:
                    raise

                else:
                    traceback.pop()
                
            r = self.result
            self._exec_level -= 1

            if self._exec_level == 0:
                self.result = None

            return r

        except BaseException as err:
            if parser:
                parser.error("({}) {}".format(type(err).__name__, str(err)))
                self.print_traceback(parser, traceback)

                if not isinstance(err, FreRuntimeError):
                    print('\nDetails:\n')
                    tback.print_exc()

            else:
                raise

    def print_traceback(self, parser, tb):
        for ind, (at, line, col) in list(enumerate(tb))[::-1]:
            if line:
                print('{}. [{}:{}] at {}:'.format(ind + 1, line, col, at))
                print('  ' + parser.lexer.input.get_line(line))
                print(' ' * (1 + col) + '^')

            else:
                print('{}. at {}'.format(ind + 1, at))

    def execute_code(self, s, tb = None):
        (parser, parse_iter) = parsing.FreParser.parse(s)
        self._parser = parser
        tb = tb or []

        return self.execute(parse_iter, tb, parser)


import functools

from . import lexing, specs


class ParserNode(object):
    def __init__(self, node_type, value = None, body = None, children = (), line = 0, column = 0):
        self.node_type = node_type
        self.value = value
        
        self.line = line
        self.column = column

        self.body = body
        self.children = tuple(children)

    def print_tree(self, level = 0):
        print('|  ' * level + '+-- Type: ' + self.node_type)

        if self.value is not None:
            if isinstance(self.value, ParserNode):
                print('|  ' * level + '+-+ Value: ')
                self.value.print_tree(level + 2)

            else:
                print('|  ' * level + '+-- Value: ' + repr(self.value))

        if self.body is not None:
            print('|  ' * level + '+--+ Body: ')
            self.body.print_tree(level + 1)
            print('|  ' * level + '|')

        if len(self.children) > 0:
            print('|  ' * level + '+--+ Children: ')
            
            for c in self.children:
                if c is None:
                    print('|  ' * (level + 1) + '+-- (none)')

                else:
                    c.print_tree(level + 2)

            print('|  ' * level + '|')

        if level > 0:
            print('|  ' * level)

class FreParser(object):
    def __init__(self, spec):
        self.lexer = lexing.FreLexer(spec)
        self.spec = spec

    def make_node(self, *args, **kwargs):
        return ParserNode(*args, line = self.lexer.input.line, column = self.lexer.input.column, **kwargs)

    def write(self, data):
        self.lexer.write(data)

    def skip_keyword(self, keyword):
        tok = self.next_any()

        if tok.type != 'keyword' or tok.value != keyword:
            self.error('Expected a keyword "{}"; got {} "{}"'.format(keyword, tok.type, tok.value))

    def is_keyword(self, keyword):
        if self.end():
            return False

        tok = self.lexer.peek()

        return tok.type == 'keyword' and tok.value == keyword

    def is_punctuation(self, char):
        if self.end():
            return False

        tok = self.lexer.peek()

        return tok.type == 'punctuation' and tok.value == char

    def expect_punctuation(self, *chars):
        if self.end():
            self.error("Expected punctuation (either of {}); got end of file instead.".format(', '.join(repr(x) for x in chars)))

        tok = self.lexer.peek()

        if not (tok.type == 'punctuation' and tok.value in chars):
            self.error("Expected punctuation (either of {}); got {} \"{}\"".format(', '.join(repr(x) for x in chars), tok.type, tok.value))

    def skip_punctuation(self, char):
        if self.end():
            self.error("Expected punctuation '{}'; got end of file instead.".format(char))

        tok = self.lexer.peek()

        if not (tok.type == 'punctuation' and tok.value == char):
            self.error("Expected punctuation '{}'; got {} \"{}\"".format(char, tok.type, tok.value))

        self.next_any()

    def skip_any(self):
        next(self.lexer)

    def next_any(self):
        return next(self.lexer)

    def error(self, msg):
        self.lexer.error(msg)

    def next_type(self):
        if self.end():
            self.error("Expected a type; got end of file instead.")

        acc = ''

        while True:
            tok = self.lexer.peek()

            if tok.type == 'type' or (tok.type == 'punctuation' and tok.value in '()'):
                acc += tok.value
                self.skip_any()
                
            elif not acc:
                self.error("Expected a type; got {} \"{}\"".format(tok.type, tok.value))

            else:
                break

        return self.make_node('type', acc)

    def next_operator(self):
        if self.end():
            self.error("Expected an operator; got end of file instead.")
        
        tok = self.next_any()

        if tok.type != 'operator':
            self.error("Expected an operator; got {} \"{}\"".format(tok.type, tok.value))
        
        return tok.value

    def next_identifier(self, can_mod = False):
        if self.end():
            self.error("Expected an identifier; got end of file instead.")

        tok = self.next_any()

        if tok.type != 'identifier':
            self.error("Expected an identifier; got {} \"{}\"".format(tok.type, tok.value))

        # Extract modifier
        mods = ''
        acc = ''
        
        for t in tok.value:
            if t in self.spec.identifier_modifiers and not acc:
                if not can_mod:
                    self.error("Identifier modifier not expected here!")

                mods += t

            else:
                acc += t

        if can_mod:
            return (acc, mods)

        else:
            return acc

    def is_type(self, *ttypes):
        if self.end():
            return False

        tok = self.lexer.peek()

        return tok.type in ttypes

    def end(self):
        return self.lexer.peek() is None

    def parse_arg_defs(self):
        self.skip_punctuation('(')

        defs = []

        while True:
            if self.is_punctuation(')'):
                self.skip_punctuation(')')
                break

            atype = self.next_type()
            aname = self.next_identifier()

            defs.append(self.make_node('expr.argdef', aname, body = atype))
            self.expect_punctuation(',', ')')

            if self.is_punctuation(','):
                self.skip_punctuation(',')

            else:
                self.skip_punctuation(')')
                break

        return defs

    def parse_func(self):
        self.skip_keyword('func')

        return_type = self.next_type()

        if self.is_type('identifier'):
            fname = self.next_identifier()

        else:
            fname = None

        children = self.subparser('func.arglist', self.parse_arg_defs)
        body = self.subparser('statement', self.parse_statement)

        node = self.make_node('expr.func', fname, return_type, (body, *children))

        return node

    def parse_args(self):
        self.skip_punctuation('(')
        res = []

        while True:
            if self.is_punctuation(')'):
                self.skip_punctuation(')')
                break

            r = self.subparser('arglist.expr', self.parse_expr)
            res.append(r)

            self.expect_punctuation(',', ')')

            if self.is_punctuation(','):
                self.skip_punctuation(',')

            else:
                self.skip_punctuation(')')
                break

        return res

    def parse_func_call(self, name):
        return self.make_node('expr.funcall', name, children = self.subparser('expr.funcall.arglist', self.parse_args))

    def parse_expr(self):
        if self.is_punctuation('('):
            self.skip_punctuation('(')
            expr = self.subparser('expr.subexpr', self.parse_expr)
            self.skip_punctuation(')')

            return expr

        tok = self.lexer.peek()

        if tok.type == 'identifier':
            name, mods = self.next_identifier(True)

            if self.is_punctuation('('):
                if mods != '':
                    self.error("Identifier modifiers not supported in function calls!")

                return self.subparser('expr.funcall', self.parse_func_call, name)

            else:
                return self.make_node('expr.identifier', (name, mods))

        elif tok.type == 'number':
            self.next_any()
            return self.make_node('expr.literal.int', tok.value)

        elif tok.type == 'number.float':
            self.next_any()
            return self.make_node('expr.literal.float', tok.value)

        elif tok.type == 'string':
            self.next_any()
            return self.make_node('expr.literal.string', tok.value)

        elif tok.type == 'value':
            self.next_any()
            return self.make_node('expr.literal.value', tok.value)

        elif tok.type == 'operator':
            return self.subparser('expr.oper', self.parse_operation)

        elif tok.type == 'keyword' and tok.value == 'func':
            return self.subparser('expr.func', self.parse_func)

        elif tok.type == 'keyword' and tok.value == 'set':
            return self.subparser('expr.set', self.parse_set)

        else:
            self.error("Expected expression; got {} \"{}\"".format(tok.type, tok.value))

    def parse_operands(self):
        res = []

        while True:
            res.append(self.subparser('expr.oper.operands.expr', self.parse_expr))
            do_next = self.is_punctuation(',')

            if do_next:
                self.skip_punctuation(',')

            else:
                break

        return res

    def parse_define_var(self):
        kind = self.next_type()
        name, mods = self.next_identifier(True)
        value = None

        if self.is_punctuation(','):
            self.skip_punctuation(',')

            if mods == '':
                value = self.subparser('var.value_expr', self.parse_expr)

            else:
                value = self.next_identifier(True)

        return self.make_node('vardef', (name, mods, value), body = kind)

    def parse_operation(self):
        operator = self.next_operator()
        operands = self.subparser('expr.oper.operands', self.parse_operands)

        return self.make_node('expr.oper', operator, children = operands)

    def parse_set(self):
        self.skip_keyword('set')
        
        var = self.next_identifier(True)
        self.skip_punctuation(',')
        val = self.subparser('set.value', self.parse_expr)

        return self.make_node('expr.set', var, body = val)

    def parse_condition(self):
        self.skip_punctuation('(')
        expr = self.subparser('condition.expr', self.parse_expr)
        self.skip_punctuation(')')

        return expr

    def parse_if(self):
        self.skip_keyword('if')
        condition = self.subparser('if.condition', self.parse_condition)
        body = self.subparser('if.body', self.parse_statement)

        if self.is_keyword('else'):
            self.skip_keyword('else')
            else_body = self.subparser('if.else', self.parse_statement)

        else:
            else_body = None

        return self.make_node('logic.if', children = (condition, self.make_node('logic.else', body = else_body)), body = body)

    def parse_while(self):
        self.skip_keyword('while')
        condition = self.subparser('while.condition', self.parse_condition)
        body = self.subparser('while.body', self.parse_statement)

        if self.is_keyword('else'):
            self.skip_keyword('else')
            else_body = self.subparser('while.else', self.parse_statement)

        else:
            else_body = None

        return self.make_node('logic.while', children = (condition, self.make_node('logic.else', body = else_body)), body = body)

    def parse_return(self):
        self.skip_keyword('return')
        value = self.subparser('return.value', self.parse_expr)
        return self.make_node('return', body = value)

    def parse_regular_statement(self):
        tok = self.lexer.peek()

        if tok.type == 'keyword':
            if tok.value == 'return':
                return self.subparser('return', self.parse_return)

            elif tok.value == 'set':
                return self.subparser('expr.set', self.parse_set)

            elif tok.value == 'break':
                self.skip_any()
                return ParserNode('break')

            elif tok.value == 'continue':
                self.skip_any()
                return ParserNode('continue')

            else:
                self.error("Expected statement; got {} \"{}\"".format(tok.type, tok.value))

        elif tok.type == 'operator':
            return self.subparser('expr.oper', self.parse_operation)

        elif tok.type == 'type':
            return self.subparser('type', self.parse_define_var)

        elif tok.type == 'identifier':
            return self.subparser('expr', self.parse_expr)

        else:
            self.error("Expected statement; got {} \"{}\"".format(tok.type, tok.value))

    def parse_statement(self):
        if self.end():
            return None

        elif self.is_punctuation(';'):
            self.skip_punctuation(';')
            return self.make_node('pass')

        elif self.is_punctuation('{'):
            children = []

            self.skip_punctuation('{')

            while True:
                if self.end():
                    self.error('Expected "}" before end of file!')

                elif self.is_punctuation('}'):
                    self.skip_punctuation('}')
                    break

                else:
                    children.append(self.subparser('statement', self.parse_statement))

            return self.make_node('block', children = children)

        else:
            tok = self.lexer.peek()

            if tok.type == 'keyword':
                if tok.value == 'func':
                    return self.subparser('expr.func', self.parse_func)

                elif tok.value == 'if':
                    return self.subparser('logic.if', self.parse_if)
                    
                elif tok.value == 'while':
                    return self.subparser('logic.while', self.parse_while)

            res = self.parse_regular_statement()
            self.skip_punctuation(';')
            return res

    def subparser(self, name, func, *args, **kwargs):
        return self.lexer.input.call(name, func, *args, **kwargs)

    @classmethod
    def parse(cls, data, spec = None):
        spec = spec or specs.FreSpecs()

        parser = cls(spec)
        parser.write(data)
        
        return (parser, iter(functools.partial(parser.subparser, 'parser', parser.parse_statement), None))

def main():
    import sys

    data = sys.stdin.read()

    for x in FreParser.parse(data, spec = specs.FreSpecs)[1]:
        x.print_tree()

    return

if __name__ == "__main__":
    main()


import string

from . import stream
from . import specs


class LexerToken(object):
    def __init__(self, token_type, value):
        self.type = token_type
        self.value = value

    def __repr__(self):
        return "({} '{}')".format(self.type, self.value)


class FreLexer(object):
    def __init__(self, lang_specs = None):
        lang_specs = lang_specs or specs.FreSpecs()
        self.input = stream.CharacterStream()
        self.lang_specs = lang_specs
    
    def write(self, data):
        self.input.write(data)

    def error(self, msg):
        self.input.error(msg)

    def read_hex_escape(self):
        data = ''

        for _ in range(2):
            ch = self.input.next()

            if ch is None:
                self.input.error("Hit end of file before proper hexadecimal character escape sequence!")

            elif ch not in '0123456789ABCDEF':
                self.input.error("Not a valid hexadecimal digit: '{}'".format(ch))

            data += ch

        return chr(int(ch, 16))

    def read_string_escape(self):
        seq = self.input.next()

        if seq is None:
            self.input.error("Hit end of file before escape sequence!")
            return ''

        elif seq == '"':
            self.input.error("Hit end of string before escape sequence!")
            return ''

        elif seq == 'n': return '\n'
        elif seq == '\\': return '\\'
        elif seq == '"': return '"'
        elif seq == "'": return "'"
        elif seq == 'x': return self.subreader('string.escape.hex', self.read_hex_escape)
        else:
            self.input.error("Hit end of file before escape sequence!")
            return ''

    def read_string(self):
        res = ""

        if self.input.peek() == "\"": self.input.next()

        while True:
            new_char = self.input.next()

            if new_char == "\"":
                break

            elif new_char == "\\":
                res += self.subreader('string.escape', self.read_string_escape).value

            else:
                res += new_char

        return LexerToken('string', res)
    
    def read_number_hex(self):
        accumulator = ''

        while True:
            if self.input.peek() not in '0123456789ABCDEF':
                break

            accumulator += self.input.next()

        if accumulator == '':
            self.input.error("No valid hexadecimal digit found! (found '{}' instead)".format(self.input.peek()))

        else:
            return int(accumulator, 16)

    def read_number_octal(self):
        accumulator = ''

        while True:
            if self.input.peek() not in '01234567':
                break

            accumulator += self.input.next()

        if accumulator == '':
            self.input.error("No valid octal digit found! (found '{}' instead)".format(self.input.peek()))

        else:
            return int(accumulator, 8)

    def read_number(self):
        accumulator = ''
        has_dot = False

        while True:
            if self.input.peek() == '.':
                if has_dot:
                    self.error("Multiple dots found in floating point literal!")

                else:
                    has_dot = True
                    accumulator += self.input.next()

            elif self.input.peek() not in string.digits:
                break

            accumulator += self.input.next()

        if has_dot:
            return LexerToken('number.float', float(accumulator))

        else:
            return int(accumulator, 10)

    def read_number_binary(self):
        accumulator = ''

        while True:
            if self.input.peek() not in '01':
                break

            accumulator += self.input.next()

        if accumulator == '':
            self.input.error("No valid binary digit found! (found '{}' instead)".format(self.input.peek()))

        else:
            return int(accumulator, 2)

    def subreader(self, name, func, *args, **kwargs):
        res = self.input.call(name, func, *args, **kwargs)

        if isinstance(res, LexerToken):
            return res

        else:
            return LexerToken(name.split('.')[0], res)

    def read_number_zero(self):
        accumulator = ""

        assert self.input.next() == '0';

        if self.input.peek() == '0': return

        if self.input.peek() in 'Xx':
            self.input.next()
            return self.subreader('number.hex', self.read_number_hex)

        elif self.input.peek() in 'Bb':
            self.input.next()
            return self.subreader('number.binary', self.read_number_binary)

        elif self.input.peek() in 'Oo':
            self.input.next()
            return self.subreader('number.octal', self.read_number_octal)

        elif self.input.peek() in string.digits + '.':
            return self.subreader('number.normal', self.read_number)

        else:
            return LexerToken('number', 0)
 
    def read_type_modifiers(self):
        accumulator = ''

        while True:
            if self.input.peek() in ' \n\t': self.input.next()
            elif self.input.peek() not in self.lang_specs.type_modifiers: break
            else: accumulator += self.input.next()

        return accumulator

    def read_identifier(self):
        accumulator = ''
        real = ''
        offset = 0

        fork = self.input.fork()

        while True:
            if fork.peek() not in self.lang_specs.identifier_modifiers: break
            accumulator += fork.next()
            offset += 1

        while True:
            if fork.peek() not in string.ascii_letters + string.digits + '_:': break
            real += fork.next()
            offset += 1

        if not real:
            return LexerToken('operator', self.read_operator())

        for _ in range(offset):
            self.input.next()

        accumulator += real

        if accumulator in self.lang_specs.keywords:
            return LexerToken('keyword', accumulator)

        elif accumulator in (x.name for x in self.lang_specs.types):
            return LexerToken('type', accumulator + self.read_type_modifiers())

        elif accumulator in self.lang_specs.values:
            return LexerToken('value', (accumulator, self.lang_specs.values[accumulator]))

        else:
            return LexerToken('identifier', accumulator)

    def read_operator(self):
        accumulator = ''

        while True:
            if not self.is_operator_char(self.input.peek()): break
            accumulator += self.input.next()

        return accumulator

    def is_operator_char(self, ch):
        return ch in '+-*/%=&|<>!@$'

    def is_punctuation(self, ch):
        ccode = ord(ch)
        
        return not (ccode <= 0x20 or ccode > 0x7E)

    def next_token(self):
        while True:
            if self.input.eof(): return

            next_char = self.input.peek()

            if next_char == '': return

            elif self.input.peek(length = 2) == '//':
                self.input.next()
                self.input.next()

                while self.input.peek() != '\n':
                    self.input.next()

            elif self.input.peek(length = 2) == '/*':
                self.input.next()
                self.input.next()
                
                while self.input.peek(length = 2) != '*/':
                    self.input.next()

            elif next_char in ' \n\t': self.input.next()
            elif next_char == '\"': yield self.subreader('string', self.read_string)
            elif next_char == '0': yield self.subreader('number.zero', self.read_number_zero)
            elif next_char in string.digits: yield self.subreader('number.normal', self.read_number)
            elif next_char in string.ascii_letters + '_:' + self.lang_specs.identifier_modifiers: yield self.subreader('identifier', self.read_identifier)
            elif self.is_operator_char(next_char): yield self.subreader('operator', self.read_operator)
            elif self.is_punctuation(next_char):
                self.input.next()
                yield LexerToken('punctuation', next_char)

            else: self.input.error("Invalid character could not be handled: '{}'".format(next_char))

    def peek(self):
        curr = (self.input.pos, self.input.line, self.input.column)

        try:
            res = next(x for x in self.next_token() if x is not None)

        except StopIteration:
            res = None

        (self.input.pos, self.input.line, self.input.column) = curr

        return res

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        return next(x for x in self.next_token() if x is not None)

    @classmethod
    def lex(cls, data):
        lexer = cls()
        lexer.write(data)

        return iter(lexer)

def main():
    import sys

    data = sys.stdin.read()

    for x in FreLexer.lex(data):
        print(repr(x))

if __name__ == "__main__":
    main()            


class FreSyntaxError(BaseException):
    pass

class CharacterStream(object):
    def __init__(self):
        self.data = ""
        self.reset_positions()
        self.routine = None
        self.errored = False

    def write(self, data):
        self.data += data
        self.lines = self.data.split('\n')

    def reset_positions(self):
        self.pos = 0
        self.line = 1
        self.column = 0

    def get_line(self, l):
        return self.lines[l - 1]

    def set_data(self, data):
        self.data = data
        self.lines = self.data.split('\n')
        self.reset_positions()

    def fork(self):
        res = CharacterStream()

        res.set_data(self.data)

        res.line = self.line
        res.column = self.column
        res.pos = self.pos

        return res

    def call(self, routine, func, *args, **kwargs):
        old_routine = self.routine
        self.routine = routine

        res = func(*args, **kwargs)

        self.routine = old_routine
        return res

    def routine_name(self):
        if self.routine is None:
            return '(unknown)'

        else:
            return self.routine

    def error(self, message = ""):
        if self.errored:
            return

        print("[{}:{} @ {}] {}".format(self.line, self.column, self.routine_name(), message))
        self.errored = True

    def next(self):
        if len(self.data) <= self.pos:
            return None

        char = self.data[self.pos]

        if char == "\n":
            self.line += 1
            self.column = 0

        else:
            self.column += 1

        self.pos += 1

        # ignore carriage return characters
        while not self.eof() and self.data[self.pos] == "\r":
            self.pos += 1

        return char

    def peek(self, offset = 0, length = 1):
        if len(self.data) <= self.pos + offset:
            return None

        return self.data[self.pos + offset: self.pos + offset + length]

    def eof(self):
        return self.pos >= len(self.data)


import os

from distutils.core import setup


with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as f:
    setup(
        name='frelang',
        description='Interpreted, statically typed, simple, flexible, embeddable language.',
        version='0.1.0-rc3',
        author='Gustavo Ramos Rehermann (Gustavo6046)',
        author_email='rehermann6046@gmail.com',
        license='The MIT License',

        packages=['fre'],
        long_description=f.read()
    )



from . import vm

import abc
import re


class LanguageType(abc.ABC):
    name = None
    default_value = None

    def __hash__(self):
        return hash(self.name)

    @abc.abstractmethod
    def check(self, obj):
        pass

    def pointer(self):
        return PointerType(self)

    def autocast(self, value):
        return value

    def __str__(self):
        return self.name

    def matches(self, other):
        return hash(self) == hash(other) or other.matches(self)

class FunctionType(LanguageType):
    def __init__(self, arg_types, return_type):
        self.arg_types = list(arg_types)
        self.return_type = return_type

        assert all(isinstance(x, LanguageType) for x in self.arg_types)
        assert isinstance(self.return_type, LanguageType)

    def __str__(self):
        return '{}({})'.format(str(self.return_type), ','.join(str(x) for x in self.arg_types))

    def check(self, obj):
        return (
            obj.return_kind.matches(self.return_type)
            and len(self.arg_types) == len(obj.arg_typenames)
            and all(
                a.matches(b)
                for (a, _), b in zip(obj.arg_typenames, self.arg_types)
            )
        )

class PointerType(LanguageType):
    name = 'pointer'
    default_value = None

    def __init__(self, deref_type):
        self.deref_type = deref_type

    def __str__(self):
        return str(self.deref_type) + '*'

    def __hash__(self):
        return hash(hash(self.deref_type) + hash('_ptr'))

    def check(self, obj):
        return obj is None or (isinstance(obj, vm.FrePointer) and obj.kind.matches(self.deref_type))

class ArrayType(LanguageType):
    name = 'array'
    default_value = ()

    def autocast(self, value):
        if not isinstance(value, list):
            return list(value)

        return value

    def __init__(self, element_type):
        self.element_type = element_type

    def __str__(self):
        return str(self.element_type) + '%'

    def __hash__(self):
        return hash(hash(self.element_type) + hash('[]'))

    def check(self, obj):
        return isinstance(obj, list) and all(self.element_type.check(o) for o in obj)

class VoidType(LanguageType):
    name = 'void'
    default_value = None

    def check(self, obj):
        return obj is None

class IntType(LanguageType):
    name = 'int'
    default_value = 0

    def check(self, obj):
        return isinstance(obj, int) or obj in (True, False)

    def autocast(self, value):
        return int(value)

class FloatType(LanguageType):
    name = 'float'
    default_value = 0.

    def check(self, obj):
        return isinstance(obj, int) or isinstance(obj, float)

    def autocast(self, value):
        return float(value)

class StringType(LanguageType):
    name = 'string'
    default_value = ''

    def check(self, obj):
        return isinstance(obj, bytes) or isinstance(obj, str)

    def autocast(self, value):
        return str(value)
class AnyType(LanguageType):
    name = 'any'
    default_value = None

    def check(self, obj):
        return True

    def matches(self, other):
        return True


##=======================##


class LanguageSpecs(object):
    keywords = ()
    types = ()
    type_modifiers = ''
    identifier_modifiers = ''
    values = {}

    def find_type(self, name):
        try:
            return next(x for x in self.types if x.name == name)

        except StopIteration:
            raise ValueError("Type '{}' not found!".format(name))

class FreSpecs(LanguageSpecs):
    keywords = ('func', 'if', 'else', 'while', 'return', 'continue', 'break', 'set')
    types = (VoidType, IntType, FloatType, StringType, AnyType)
    type_modifiers = '*%'
    identifier_modifiers = '@*'
    values = {'none': None, 'null': None, 'true': True, 'false': False}